# $Id: ILUTestingMachinery.py,v 1.8 1999/01/22 03:23:24 janssen Exp $
# BeginILUCopyright

# Copyright (c) 1991-1998 Xerox Corporation.  All Rights Reserved.

# Unlimited use, reproduction, modification, and distribution of this
# software and modified versions thereof is permitted.  Permission is
# granted to make derivative works from this software or a modified
# version thereof.  Any copy of this software, a modified version
# thereof, or a derivative work must include both the above copyright
# notice of Xerox Corporation and this paragraph.  Any distribution of
# this software, a modified version thereof, or a derivative work must
# comply with all applicable United States export control laws.  This
# software is made available AS IS, and XEROX CORPORATION DISCLAIMS ALL
# WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE, AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY
# LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
# EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
# NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGES.

# EndILUCopyright

import sys, re, os

SBHPattern = re.compile('(ilusbh:|IOR:|ilu:)([a-zA-Z0-9@/:=+-._;%]*)')

from ILUTestingConfig import iluconf_dict, makefile_dict, iludefs_dict

#
# Some exit wizardry to kill servers, delete log files, etc., automagically
#

def do_exit_function (fn, next_fn):
	fn()
	if next_fn:
		next_fn()

def add_exit_function (fn):
	if sys.__dict__.has_key('exitfunc'):
		oldexit = sys.exitfunc
	else:
		oldexit = None
	sys.exitfunc = lambda x=oldexit, y=fn: do_exit_function(y,x)

def killproc(pid):
	import os
	try:
		os.kill(pid, 9)
	except:
		pass

def delfile(fname):
	import os
	try:
		os.unlink(fname)
	except:
		pass

# A msg function which only writes when ILUTEST_VERBOSE is set

Verbose = os.environ.has_key('ILUTEST_VERBOSE')

def user_msg (msg):
	global Verbose
	if Verbose:
		sys.stderr.write(msg)

# A function which adds to an environment variable

def prefix_environment_variable (var_name, value):
	existing_value = (os.environ.has_key(var_name) and os.environ[var_name])
	if existing_value:
		new_value = value + ":" + existing_value
	else:
		new_value = value
	os.environ[var_name] = new_value

# Exceptions to raise when process oddities are encountered

class Already_Started:
	def __init__ (self, proc):
		self.proc = proc

class Server_Failed:
	def __init__ (self, server_commands):
		self.commands = server_commands

	def __str__(self):
		import string
		return '<Server_Failed:  ' + string.join(server_commands) + '>'

	__repr__ = __str__

class Error_Exit_Status:
	def __init__ (self, status, program, logfile):
		self.status = status
		self.program = program
		self.logfile = logfile

	def __str__(self):
		return '<Error_Exit_Status:  status=%s, program="%s", logfile="%s">' % (self.status, self.program, self.logfile)

	__repr__ = __str__

class TestError:
	pass

# Logfile class to keep output in

class Logfile:
	def __init__ (self, name=None):
		import tempfile
		if not name:
			self.filename = tempfile.mktemp()
		else:
			self.filename = name
		self.verbose = Verbose
		self.os = os
		add_exit_function(lambda x=self.filename: delfile(x))

	def display (self, banner=None):
		if banner:
			print banner
		f = open(self.filename,'r')
		line = f.readline()
		while line:
			sys.stdout.write(line)
			line = f.readline()
		f.close()

	def exists(self):
		return os.path.exists(self.filename)

	def open(self, mode='r'):
		return open(self.filename, mode)

	def delete (self):
		import os
		if os.path.exists(self.filename):
			os.unlink(self.filename)
		
	zero = delete

	def __str__(self):
		return '<Logfile %s>' % self.filename

	__repr__ = __str__

# A Process class that wraps a running server or client process (or any other kind of process)

def find_executable (program_name, opt_path=None):

	def is_executable (path):
		import os
		mode = os.stat(path)[0]
		return ((mode & 0111) != 0)

	import string, os
	pname = os.path.expanduser(os.path.expandvars(program_name))
	if pname[0] == '.' or pname[0] == '/':
		return program_name
	path = opt_path or os.environ["PATH"]
	dirs = string.split(path, os.pathsep)
	for dir in dirs:
		test_path = os.path.join(dir, pname)
		if os.path.exists(test_path) and os.path.isfile(test_path) and is_executable(test_path):
			return test_path
	return None

class Process:
	def __init__ (self, executable, args, env, logfile=None):
		import tempfile
		import os
		self.logfile = logfile
		self.executable = find_executable(executable)
		self.args = args
		self.exitstatus = None
		self.running = 0
		self.pid = os.fork()
		if self.pid == 0:	# child process
			tempfile.template = None
			if logfile:
				fd = os.open(logfile.filename, os.O_WRONLY | os.O_CREAT | os.O_APPEND, 0664)
				os.close(1)
				os.dup2(fd, 1)
				os.close(2)
				os.dup2(fd, 2)
			for key in env.keys():
				os.environ[key] = env[key]
			sys.stderr.write("executable is " + str(self.executable) + ", args are " + str(args) + ", env = " + str(os.environ) + "\n");
			os.execve(self.executable, (self.executable,) + args, os.environ)
		else:
			self.running = 1
			add_exit_function(lambda x=self.pid: killproc(x))

	def active(self):
		import os
		if not self.running:
			return 0
		status = os.waitpid (self.pid, os.WNOHANG)
		if (status[0] == 0):
			# process still running
			return 1
		else:
			self.exitstatus = status[1]/256
			self.running = 0
			return 0

	def finish (self):
		import os
		if self.running:
			status = os.waitpid (self.pid, 0)
			self.exitstatus = status[1]/256
			self.running = 0
			return status[1]%256
		else:
			return 0

	def kill (self, signal=9):
		import posix
		if self.running:
			posix.kill(self.pid, signal)
			return self.finish()
		else:
			return 0
		
class Program:

	program_type = "Program"

	def __init__ (self, executable, default_args=(), env={}):
		import types
		if type(default_args) == types.StringType:
			default_args = ( default_args, )
		if type(default_args) != types.TupleType:
			default_args = tuple(default_args)
		self.executable = executable
		self.default_args = default_args
		self.proc = None
		self.env_bindings = env

	def start (self, logfile, args=()):
		import types
		if self.proc:
			raise Already_Started(self.proc)
		if not args:
			args = self.default_args
		if (type(args) != types.TupleType):
			args = tuple(args)
		user_msg ("starting " + self.program_type +
			  " " + str(self.executable) + " with args '" +
			  str(args) + "' and output to " + str(logfile) + "\n")
		self.proc = Process (self.executable, args, self.env_bindings, logfile)

	def finish (self):
		status = self.proc.finish()
		exitstatus = self.proc.exitstatus
		self.proc = None
		return status, exitstatus

	def run (self, logfile):
		self.start(logfile)
		killsignal, exitstatus = self.finish()
		if exitstatus != 0:
			raise Error_Exit_Status (exitstatus, self, logfile)

	def active (self):
		# returns TRUE if program is still running
		return self.proc and self.proc.active()

	def kill (self, signal=9):
		status = self.proc.kill(signal)
		exitstatus = self.proc.exitstatus
		self.proc = None
		return status, exitstatus

	def logfile (self):
		if not self.proc:
			raise Not_Started (self)
		return self.proc.logfile.open('r')

	def __str__(self):
		import string
		return '<%s.%s "%s" proc=%s>' % (self.__class__.__module__, self.__class__.__name__,
						  string.join((self.executable,) + self.default_args),
						  self.proc and self.pid)

	__repr__ = __str__

class Server (Program):
	program_type = "Server"

class SBH_Server (Server):
	"""subclass of Server which writes its SBH to stdout.  The SBH can be retrieved as server.sbh"""
	def __init__(self, executable, default_args=(), env={}):
		Server.__init__(self, executable, default_args, env)
		self.sbh = None

	def start (self, logfile, args=()):
		import time
		global SBHPattern
		Server.start (self, logfile, args)
		while not logfile.exists():
			time.sleep(1)	# wait for file to be created
		time.sleep(10)		# wait for service to be registered
		f = logfile.open('r')
		junkline = f.readline()
		sbhline = f.readline()
		f.close()
		sbhmatch = SBHPattern.search(sbhline)
		if sbhmatch:
			self.sbh = sbhline[sbhmatch.start(1):sbhmatch.end(2)]

class Client (Program):
	program_type = "Client"

class Test:

	def normalize_command (self, cmd, cmd_name):
		import types, string
		if type(cmd) == types.StringType:
			c = string.split(cmd)
		elif (type(cmd) == types.TupleType or
			 type(cmd) == types.SequenceType):
			c = cmd
		else:
			raise "bad " + cmd_name +" '" + str(cmd) + "' supplied."
		return c

class SimpleClientServerTest(Test):

	def __init__(self, test_name, client_commands, server_commands):
		self.name = test_name
		self.client_commands = self.normalize_command(client_commands, "client commands")
		self.server_commands = self.normalize_command(server_commands, "server commands")
		self.server_startup_delay = 5

	def run(self):
		import time, traceback, string

		oldpath = os.environ["PATH"]
		os.environ["PATH"] = "." + os.pathsep + os.environ["PATH"]

		print 'Running test "' + self.name + '"...'

		clientlog = None
		serverlog = None

		try:
			try:

				serverlog = Logfile()

				# start server -- sys.executable holds pathname of Python interpreter
				s = Server (self.server_commands[0], self.server_commands[1:])
				s.start(serverlog)
				time.sleep(self.server_startup_delay)	# let it get started
				if not s.active():
					raise 'server startup failed : ' + string.join(self.server_commands)
			
				# run client against server
				clientlog = Logfile()
				Client (self.client_commands[0], self.client_commands[1:]).run(clientlog)
			
				# OK, things are fine.  Print a message and exit
				print 'Test "' + self.name + '" completed successfully.'

				# kill the server
				s.kill()

			except:
				print "Error in test " + self.name + ":"
				traceback.print_exc()
				if serverlog:
					serverlog.display("******* Server Log ***********")
				if clientlog:
					clientlog.display("******* Client Log ***********")
				raise TestError()

		finally:
			os.environ["PATH"] = oldpath


class ClientSBHServerTest(Test):
	"""Subclass of Test in which the server writes its SBH or IOR to stdout, and the client uses that SBH
	in one of its command-line arguments.  The string '$SBH$' in any client command-line argument will be
	replaced with the actual SBH output by the server."""

	def __init__(self, test_name, client_commands, server_commands):
		self.name = test_name
		self.client_commands = self.normalize_command(client_commands, "client commands")
		self.server_commands = self.normalize_command(server_commands, "server commands")
		self.server_startup_delay = 5

	def run(self):
		import time, traceback, string

		oldpath = os.environ["PATH"]
		os.environ["PATH"] = "." + os.pathsep + os.environ["PATH"]

		print 'Running test "' + self.name + '"...'

		clientlog = None
		serverlog = None

		try:
			try:

				serverlog = Logfile()

				# start server -- sys.executable holds pathname of Python interpreter
				s = SBH_Server (self.server_commands[0], self.server_commands[1:])
				s.start(serverlog)
				time.sleep(self.server_startup_delay)	# let it get started
				if not s.active():
					raise 'server startup failed : ' + string.join(self.server_commands)
				if not s.sbh:
					raise "can't find server SBH"
			
				# run client against server
				clientlog = Logfile()
				commands = []
				for command in self.client_commands:
					real_command = string.replace(command, '$SBH$', s.sbh);
					commands.append(real_command)
				Client (commands[0], commands[1:]).run(clientlog)
			
				# OK, things are fine.  Print a message and exit
				print 'Test "' + self.name + '" completed successfully.'

				# kill the server
				s.kill()

			except:
				print "Error in test " + self.name + ":"
				traceback.print_exc()
				if serverlog:
					serverlog.display("******* Server Log ***********")
				if clientlog:
					clientlog.display("******* Client Log ***********")
				raise TestError()

		finally:
			os.environ["PATH"] = oldpath


class SimpleSingleProgramTest (Test):

	def __init__(self, test_name, commands):
		import types, string
		self.name = test_name
		self.commands = self.normalize_command(commands, "commands")

	def run(self):
		import time, traceback

		oldpath = os.environ["PATH"]
		os.environ["PATH"] = "." + os.pathsep + os.environ["PATH"]

		print 'Running test "' + self.name + '"...'

		try:
			try:

				logfile = Logfile()

				s = Program(self.commands[0], self.commands[1:])
				s.run(logfile)
			
				# OK, things are fine.  Print a message and exit
				print 'Test "' + self.name + '" completed successfully.'

			except:
				print "Error in test " + self.name + ":"
				traceback.print_exc()
				logfile.display("******* Program Log ***********")
				raise TestError()

		finally:
			os.environ["PATH"] = oldpath

